
/*
 * par: simple paragraph formater
 *
 * (C) 1998 by Kamil Toman (e-mail: toman@horac.ta.jcu.cz)
 *
 * 24th April 1999
 * version: 1.1
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licences as by published
 * by the Free Software Foundation; either version 2; or (at your option)
 * any later version
 *
 * This program is distributed in the hope that it will entertaining,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILTY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Publis License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>

#include "interfac.h"
#include "hashtab.h"
#include "par.h"

static unsigned char *string;
static int *sprice, *bprice, *length, *start, *bestprice, *eol, **prev;
static unsigned int maxpos;
static size_t strbuffsize = STRBUF, numbuffsize = NUMBUF;

static INLINE int breakprice(int startpos, int endpos, int llength)
{
    int spaces, sum;
    int linelength = length[endpos] - length[startpos] + endpos - startpos - 1;

    if (linelength > llength || bprice[endpos] == INT_MAX)
	return INT_MAX;
    if (endpos == maxpos)
	return 0;

    sum = sprice[endpos - 1] - sprice[startpos];
    spaces = llength - linelength;

    if (spaces && !sum)
	return INT_MAX;

    if (sum)
	sum = spaces * FIXED / sum;

    return (sum + spaces * spaces + bprice[endpos]);
}


static int findbreaks(int llength, int parindent, int hangindent)
{
    int totlines, i, tmp, price, prevpos, pos, startpos;
    int line = 0, tline = 0;
    int bestp, bestprev = 0;
    int tprev = 0, tbest = INT_MAX;
    int prevbegin, prevend, begin = 0, end = 1;
    int again = 1;

    ALLOC(bestprice, numbuffsize);
    ALLOC(start, numbuffsize);
    ALLOC(eol, numbuffsize);
    ALLOC(prev, numbuffsize);

    bestprice[0] = 0;
    for (pos = 1; pos <= maxpos; pos++)
	bestprice[pos] = INT_MAX;

    while (line < maxpos && again) {
	again = 0;
	prevbegin = begin, prevend = end;

	while (bestprice[begin] == INT_MAX && begin < end)
	    begin++;
	if (begin == end) {
	    line--;
	    break;
	}
	tmp = length[end - 1];
	while (length[end] - tmp <= llength && end < maxpos)
	    end++;

	ALLOC(prev[line], (end - begin + 1) * sizeof(int));
	start[line] = begin;

	startpos = prevend;
	for (pos = end; pos >= begin; pos--) {
	    bestp = INT_MAX;
	    tmp = length[pos];
	    while (tmp - length[startpos] < llength && startpos > prevbegin)
		startpos--;

	    tmp = MIN(pos, prevend);
	    for (prevpos = startpos; prevpos < tmp; prevpos++) {
		if (bestprice[prevpos] <= INT_MAX / 2 && (price = breakprice(prevpos, pos, llength)) != INT_MAX)
		    price += bestprice[prevpos];
		else
		    price = INT_MAX;
		if (price < bestp)
		    bestp = price, bestprev = prevpos;
	    }
	    bestprice[pos] = bestp;
	    if (bestp < tbest)
		again = 1;
	    prev[line][pos - begin] = bestprev;
	    if (pos == maxpos && bestp < tbest)
		tbest = bestp, tprev = pos, tline = line;
	}
	line++;
	if (line == hangindent)
	    llength += parindent;
    }
    if (tbest == INT_MAX)
	err_exit(EXIT_FAILURE, ERRNOPOS);

    totlines = tline + 1;
    while (tline > 0)
	eol[tline] = tprev, tprev = prev[tline][tprev - start[tline]], tline--;
    eol[tline] = tprev;

    for (i = 0; i < line; i++)
	free(prev[i]);
    free(prev);
    free(start);
    free(bestprice);
    free(bprice);

    return totlines;
}


static void compute_sprices(void)
{
    int i, lastsum = 0, sum = 0;


    ALLOC(sprice, numbuffsize);
    sprice[0] = 0;
    for (i = 1; i < maxpos; i++) {
	int l = length[i] - length[i - 1];
	int p = length[i + 1] - length[i];
	l = MIN(l, 8);
	p = MIN(p, 8);
	sum += sqrt(MIN(pow((float) l, 1.5), pow((float) p, 1.5)) * 3 + sqrt((float) (l + p))) * 10;
	sprice[i] = sum;
	if (sum < lastsum)
	    err_exit(EXIT_FAILURE, ERRTOOLARGE);
	lastsum = sum;
    }
    sprice[i] = sum;
}


static INLINE void compute_bprice(int i, int size, int pos)
{
    unsigned char *tmp;

    ALLOC(tmp, (size + 1) * sizeof(unsigned char));
    memcpy(tmp, string + length[i - 1], size);
    tmp[size++] = 0;

    if (is_in_hash(tmp, size))
	bprice[maxpos] = INT_MAX;
    else
	switch (string[pos]) {
	case ',':
	case ';':
	    bprice[i] = 0;
	    break;
	case '?':
	case '!':
	    bprice[i] = 10;
	    break;
	case ')':
	    bprice[i] = 20;
	    break;
	case '"':
	case '\'':
	    bprice[i] = 25;
	    break;
	case '.':
	    bprice[i] = 30;
	    break;
	default:
	    bprice[i] = 60;
	}

    free(tmp);
}


static void output(int startpos, int endpos, int indent, int llength)
{
    int i, j, spaces = 0, pos = endpos - startpos - 1;
    int *oldabsprices, *absprices, *newspaces;
    int size = MAX(pos + 1, 1);

    if (endpos != maxpos)
	spaces = llength - length[endpos] + length[startpos] - pos;

    ALLOC(oldabsprices, size * sizeof(int));
    ALLOC(absprices, size * sizeof(int));
    if ((newspaces = calloc(size, sizeof(int))) == NULL)
	 err_exit(EXIT_FAILURE, ERRNOMEM);

    for (i = startpos; i < endpos - 1; i++)
	oldabsprices[i - startpos] = absprices[i - startpos] = (sprice[i + 1] - sprice[i]) * 256 + rand() % 16;

    while (spaces) {
	int besti = 0, bestp = 0;

	for (i = 0; i < pos; i++)
	    if (absprices[i] > bestp)
		besti = i, bestp = absprices[i];

	absprices[besti] = oldabsprices[besti] / (newspaces[besti] * 2 + 3);
	newspaces[besti]++;
	spaces--;
    }

    for (i = 0; i < indent; i++)
	putc(' ', stdout);
    for (i = startpos; i < endpos - 1; i++) {
	for (j = length[i]; j < length[i + 1]; j++)
	    putc(string[j], stdout);
	for (j = 0; j <= newspaces[i - startpos]; j++)
	    putc(' ', stdout);
    }
    for (j = length[i]; j < length[endpos]; j++)
	putc(string[j], stdout);

    putc('\n', stdout);

    free(oldabsprices);
    free(absprices);
    free(newspaces);
}


static void input(void)
{
    int c, lastp = 0, p = 0, numbuff = numbuffsize / (sizeof(int)), strbuff = strbuffsize / (sizeof(char));

    maxpos = 0;
    length[0] = 0;
    while (1) {
	while ((c = getc(stdin)) == '\n' || c == ' ' || c == '\t') ;
	ungetc(c, stdin);
	while ((c = getc(stdin)) != '\n' && c != ' ' && c != '\t' && c != EOF) {
	    if (p >= strbuff) {
		strbuffsize += STRBUF;
		REALLOC(string, strbuffsize);
		strbuff = strbuffsize / (sizeof(char));
	    }
	    string[p++] = c;
	    if (p < lastp)
		err_exit(EXIT_FAILURE, ERRTOOLARGE);
	    lastp = p;
	}
	if (c == EOF)
	    break;

	if (++maxpos >= numbuff) {
	    numbuffsize += NUMBUF;
	    REALLOC(length, numbuffsize);
	    REALLOC(bprice, numbuffsize);
	    numbuff = numbuffsize / (sizeof(int));
	}
	length[maxpos] = p;
	compute_bprice(maxpos, length[maxpos] - length[maxpos - 1], p - 1);
    }
    bprice[0] = INT_MAX;

    remove_hash_table();

    strbuffsize = p * sizeof(unsigned char);
    numbuffsize = (maxpos + 1) * sizeof(int);

    REALLOC(string, strbuffsize);
    REALLOC(length, numbuffsize);
    REALLOC(bprice, numbuffsize);
}


int main(int argc, char *argv[])
{
    int i, totlines, param[NUMOFPARAMS];

    parse_cmdline_parameters(argc, argv, param);

    ALLOC(string, strbuffsize);
    ALLOC(length, numbuffsize);
    ALLOC(bprice, numbuffsize);

    input_words((MAX(RIGHT, RIGHT + PARINDENT) + 1));
    input();
    compute_sprices();
    totlines = findbreaks(RIGHT, PARINDENT, HANGINDENT);
    output(0, eol[0], LEFT, RIGHT);
    for (i = 1; i < totlines; i++) {
	if (i == HANGINDENT)
	    LEFT -= PARINDENT, RIGHT += PARINDENT;
	output(eol[i - 1], eol[i], LEFT, RIGHT);
    }

    free(length);
    free(eol);
    free(sprice);
    free(string);

    return EXIT_SUCCESS;
}
